try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type:ignore
from typing import Sequence
from pydantic import Field
from fhirkit.elements import (
    Identifier,
    Reference
)
from fhirkit.Resource import DomainResource
from fhirkit.elements.elements import Identifier


class Coverage(DomainResource):
    identifier: Sequence[Identifier] = Field(
        None,
        title="Business identifier(s) for this coverage")
    status: Literal["active","cancelled","draft","entered-in-error"] = Field(
        None,
        title="active | cancelled | draft | entered-in-error",
        valueset="https://build.fhir.org/valueset-fm-status.html")
    kind: Literal["insurance","self-pay","other"] = Field(
        None,
        title="insurance | self-pay | other",
        valueset="https://build.fhir.org/valueset-coverage-kind.html")
    beneficiary: Reference = Field(
        None,
        enum_reference_types=["Patient"],
        title="Plan beneficiary")    





try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore
from pydantic import Field
from typing import Optional, Sequence
from fhirkit import Reference
from fhirkit.Resource import DomainResource
from fhirkit.elements import (
    dateTime,
    HumanName,
    Identifier,
    ContactPoint,
    Address,
    AdministrativeGender,
    Attachment,
    Period,
    BackboneElement,
    CodeableConcept,
    DaysOfWeek,
)

class AvailableTime(BackboneElement):
    daysOfWeek: Optional[Sequence[DaysOfWeek]] = Field(
        None,
        title="mon | tue | wed | thu | fri | sat | sun")
    allDay: Optional[bool] = Field(
        None,
        title="Always available? e.g. 24 hour service")
    availableStartTime: Optional[dateTime] = Field(
        None,
        title="Opening time of day (ignored if allDay = true)")
    availableEndTime: Optional[dateTime] = Field(
        None,
        title="Closing time of day (ignored if allDay = true)")


class NotAvailable(BackboneElement):
    description: str = Field(
        None,
        title="Reason presented to the user explaining why time not available")
    during: Optional[Period] = Field(
        None,
        title="Service not available from this date")

class PractitionerRole(DomainResource):
    resourceType: Literal["PractitionerRole"] = Field(
        "PractitionerRole", 
        const=True)
    identifier: Optional[Sequence[Identifier]] = Field(
        None,
        title="An identifier for this patient")
    active: Optional[bool] = Field(
        None,
        title="Whether this practitioner's record is in active use")
    period: Optional[Period] = Field(
        None,
        title="The period during which the practitioner is authorized to perform in these role(s)")
    practitioner: Optional[Reference] = Field(
        None,
        enum_reference_types=["Practitioner"],
        title="Practitioner that is able to provide the defined services for the organization")
    organization: Optional[Reference] = Field(
        None,
        enum_reference_types=["Organization"],
        title="Organization where the roles are available")
    code: Optional[Sequence[CodeableConcept]] = Field(
        None,
        title="Roles which this practitioner may perform",
        valueset="http://hl7.org/fhir/ValueSet/practitioner-role")
    specialty: Optional[Sequence[CodeableConcept]] = Field(
        None,
        title="Specific specialty of the practitioner",
        valueset="http://hl7.org/fhir/ValueSet/c80-practice-codes")
    location: Optional[Sequence[Reference]] = Field(
        None,
        enum_reference_types=["Location"],
        title="The location(s) at which this practitioner provides care")
    healthcareService: Optional[Sequence[Reference]] = Field(
        None,
        enum_reference_types=["HealthcareService"],
        title="The list of healthcare services that this worker provides for this role's Organization/Location(s)")
    telecom: Optional[Sequence[ContactPoint]] = Field(
        None,
        title="Contact details that are specific to the role/location/service")
    availableTime: Optional[Sequence[AvailableTime]] = Field(
        None,
        title="Times the Service Site is available")
    notAvailable: Optional[Sequence[NotAvailable]] = Field(
        None,
        title="Not available during this time due to provided reason")
    availabilityExceptions: Optional[str] = Field(
        None,
        title="Description of availability exceptions")
    endpoint: Optional[Sequence[Reference]] = Field(
        None,
        enum_reference_types=["Endpoint"],
        title="Technical endpoints providing access to services operated for the practitioner with this role")
    
    

    
    
    



from __future__ import annotations
from datetime import datetime, date, time
from pathlib import Path
import re

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated # type: ignore

from typing import Optional, Union, no_type_check
from pydantic import ConstrainedStr, AnyUrl, Field


class Id(ConstrainedStr):
    regex = re.compile(r"[A-Za-z0-9\-\.]{1,64}")


class Code(ConstrainedStr):
    regex = re.compile(r"[^\s]+(\s[^\s]+)*")


class Instant(ConstrainedStr):
    regex = re.compile(
        r"([0-9]([0-9]([0-9][1-9]|[1-9]0)|[1-9]00)|[1-9]000)-(0[1-9]|1[0-2])-(0[1-9]|[1-2][0-9]|3[0-1])T([01][0-9]|2[0-3]):[0-5][0-9]:([0-5][0-9]|60)(\.[0-9]+)?(Z|(\+|-)((0[0-9]|1[0-3]):[0-5][0-9]|14:00))"
    )


class XHTML(ConstrainedStr):
    pass


class dateTime(datetime):
    pass


class URN(ConstrainedStr):
    # TODO extract nid and nss
    regex = re.compile(r"urn:\S*")


class UUID(ConstrainedStr):
    regex = re.compile(
        r"urn:uuid:[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"
    )


class URI(ConstrainedStr):
    regex = re.compile(r"\S*")


class AbsoluteURL(AnyUrl):
    pass


class RelativeURL(ConstrainedStr):
    regex = re.compile(
        r"(Account|ActivityDefinition|AdministrableProductDefinition|AdverseEvent|AllergyIntolerance|Appointment|AppointmentResponse|AuditEvent|Basic|Binary|BiologicallyDerivedProduct|BodyStructure|Bundle|CapabilityStatement|CarePlan|CareTeam|CatalogEntry|ChargeItem|ChargeItemDefinition|Citation|Claim|ClaimResponse|ClinicalImpression|ClinicalUseDefinition|CodeSystem|Communication|CommunicationRequest|CompartmentDefinition|Composition|ConceptMap|Condition|Consent|Contract|Coverage|CoverageEligibilityRequest|CoverageEligibilityResponse|DetectedIssue|Device|DeviceDefinition|DeviceMetric|DeviceRequest|DeviceUseStatement|DiagnosticReport|DocumentManifest|DocumentReference|Encounter|Endpoint|EnrollmentRequest|EnrollmentResponse|EpisodeOfCare|EventDefinition|Evidence|EvidenceReport|EvidenceVariable|ExampleScenario|ExplanationOfBenefit|FamilyMemberHistory|Flag|Goal|GraphDefinition|Group|GuidanceResponse|HealthcareService|ImagingStudy|Immunization|ImmunizationEvaluation|ImmunizationRecommendation|ImplementationGuide|Ingredient|InsurancePlan|Invoice|Library|Linkage|List|Location|ManufacturedItemDefinition|Measure|MeasureReport|Media|Medication|MedicationAdministration|MedicationDispense|MedicationKnowledge|MedicationRequest|MedicationStatement|MedicinalProductDefinition|MessageDefinition|MessageHeader|MolecularSequence|NamingSystem|NutritionOrder|NutritionProduct|Observation|ObservationDefinition|OperationDefinition|OperationOutcome|Organization|OrganizationAffiliation|PackagedProductDefinition|Patient|PaymentNotice|PaymentReconciliation|Person|PlanDefinition|Practitioner|PractitionerRole|Procedure|Provenance|Questionnaire|QuestionnaireResponse|RegulatedAuthorization|RelatedPerson|RequestGroup|ResearchDefinition|ResearchElementDefinition|ResearchStudy|ResearchSubject|RiskAssessment|Schedule|SearchParameter|ServiceRequest|Slot|Specimen|SpecimenDefinition|StructureDefinition|StructureMap|Subscription|SubscriptionStatus|SubscriptionTopic|Substance|SubstanceDefinition|SupplyDelivery|SupplyRequest|Task|TerminologyCapabilities|TestReport|TestScript|ValueSet|VerificationResult|VisionPrescription)\/[A-Za-z0-9\-\.]{1,64}(\/_history\/[A-Za-z0-9\-\.]{1,64})?"
    )

    @no_type_check
    def __new__(cls, url: ConstrainedStr, **kwargs) -> object:
        return str.__new__(cls, url)

    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate
        yield cls.validate_relative_path

    def __init__(
        self,
        url: ConstrainedStr,
        *,
        resourceType: str,
        resourceId: str,
        version: Optional[str] = None,
    ) -> None:
        ConstrainedStr.__init__(url)
        self.resourceType = resourceType
        self.resourceId = resourceId
        self.version = version

    @classmethod
    def validate_relative_path(cls, v: ConstrainedStr):
        resourceType, *other = v.split("/")
        if len(other) == 1:
            resourceId = other[0]
            version = None
        else:
            resourceId, _, version = other
        return cls(v, resourceType=resourceType, resourceId=resourceId, version=version)


literal = Annotated[Union[AbsoluteURL, RelativeURL, URN], Field()]


class CanonicalURL(AnyUrl):
    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate
        yield cls.validate_version

    def __init__(
        self,
        url: str,
        *,
        scheme: str,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        tld: Optional[str] = None,
        host_type: str = "domain",
        port: Optional[str] = None,
        path: Optional[str] = None,
        query: Optional[str] = None,
        fragment: Optional[str] = None,
        version: Optional[str] = None,
    ) -> None:
        self.version = version
        super().__init__(
            url,
            scheme=scheme,
            user=user,
            password=password,
            host=host,
            tld=tld,
            host_type=host_type,
            port=port,
            path=path,
            query=query,
            fragment=fragment,
        )

    @classmethod
    def validate_version(cls, v: CanonicalURL):
        # version suffixes are always part of a path
        if v.path is None:
            return v
        # extract the version suffix
        parts = v.path.split("|")
        if len(parts) > 1:
            v.path = parts[0]
            v.version = parts[1]
        return v

    @property
    def uri(self):
        kwargs = dict()
        kwargs["scheme"] = self.scheme
        kwargs["user"] = self.user
        kwargs["password"] = self.password
        if self.host is not None:
            kwargs["host"] = self.host

        kwargs["port"] = self.port
        kwargs["path"] = self.path
        kwargs["query"] = self.query
        kwargs["fragment"] = self.fragment
        return self.build(**kwargs)


class CanonicalURN(URN):
    version = None

    @property
    def uri(self):
        return str(self)

    # TODO extract version


canonical = Annotated[
    Union[CanonicalURL, CanonicalURN],
    Field(
        description="""A reference to a Canonical Resources (resources with `url` element) based on the canonical URL.
    They are absolute URLs or fragment identifiers with optionally a `|` version suffix.

    If no version is specified, the server takes the last version
    Fragments refer to contained resources. More info about the use of fragments: https://www.hl7.org/fhir/references.html#canonical-fragments
    """
    ),
]


class decimal(float):
    pass



from typing import Optional, Sequence
from pydantic import Field
from fhirkit.elements import ContactPoint, Element, CodeableConcept, HumanName,Address,AdministrativeGender,Reference,Period


class ContactDetail(Element):
    relationship: Optional[Sequence[CodeableConcept]] = Field(
        None, 
        title="The kind of relationship")
    name: Optional[HumanName]= Field(
        None,
        title="A name associated with the contact")
    telecom: Optional[Sequence[ContactPoint]] = Field(
        None,
        title="A contact detail for the person")
    address: Optional[Address] = Field(
        None,
        title="Address for the contact person")
    gender: Optional[AdministrativeGender] = Field(
        None,
        title="male | female | other | unknown")
    organization: Optional[Reference] = Field(
        None,
        enum_reference_types=["Organization"],
        title="Organization that is associated with the contact")
    period: Optional[Period] = Field(
        None,
        title="Time period when the contact detail is applicable")




from datetime import datetime  # this is important to have at the top

from typing import List, Optional, Sequence, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

from pydantic import Field
from fhirkit.choice_type import ChoiceType
from fhirkit.Resource import DomainResource
from fhirkit.elements import (
    CodeableConcept,
    Identifier,
    Period,
    Reference,
    Annotation,
    BackboneElement
)

ClinicalImpressionStatus = Literal[
    "in-progress",
    "completed",
    "entered-in-error"
]

class ClinicalImpressionFinding(BackboneElement):
    item: Optional[Reference] = ChoiceType(None)
    cause: Optional[str] = Field(None)  

class ClinicalImpression(DomainResource):
    resourceType: Literal["ClinicalImpression"] = Field("ClinicalImpression", const=True)
    identifier: Optional[List[Identifier]] = Field(None)
    status: Optional[ClinicalImpressionStatus] = Field(None)
    code: Optional[CodeableConcept] = Field(None)
    description: Optional[str] = Field(None)
    subject: Optional[Reference] = Field(None)
    encounter: Optional[Reference] = Field(None)
    effectiveDateTime: Optional[datetime] = Field(None)
    effectivePeriod: Optional[Period] = Field(None)
    effective: Optional[Union[datetime, Period,str]] = ChoiceType(None)
    date: Optional[datetime] = Field(None)
    assessor: Optional[Reference] = Field(None)
    previous: Optional[Reference] = Field(None)
    problem: Optional[List[Reference]] = Field(None)
    prognosisCodeableConcept: Optional[List[CodeableConcept]] = Field(None)
    supportingInfo: Optional[List[Reference]] = Field(None)
    note: Optional[List[Annotation]] = Field(None)
    finding: Optional[List[ClinicalImpressionFinding]] = Field(None)
    
    





from typing import Type, TypeVar
import warnings
from pydantic import BaseModel

TModel = TypeVar("TModel", bound=BaseModel)
def deprecated(*, reason: str):
    """This is a decorator which can be used to mark Pydantic models as deprecated."""
    def decorator(cls: Type[TModel]):
        original_init = cls.__init__
        original_doc = cls.__doc__
        def new_init(self, *args, **kwargs):
            warnings.warn("This class is deprecated: " + reason, DeprecationWarning, stacklevel=2)
            original_init(self, *args, **kwargs)
        cls.__init__ = new_init

        cls.__doc__ = "Deprecated: " + reason
        if original_doc is not None:
            cls.__doc__ += "\n\n" + original_doc
        return cls
    return decorator


from .Resource import Resource
from .elements import (
    CodeableConcept, 
    Coding, 
    Quantity, 
    Reference, 
    Identifier,
    Period,Range,
    Annotation,
    ContactPoint,
    HumanName,
    Address,
    Attachment,
    Signature,
    Meta,
    Extension,
    UsageContext,
    Duration,
    Annotation,
    UsageContext
)
from .CodeSystem import CodeSystem
from .ValueSet import ValueSet, SimpleValueSet
from .snomed import SCTCoding, SCTConcept, SCTFHIRTerminologyServer
from .tiro import TiroCoding
from .Observation import Observation
from .Procedure import Procedure, ProcedurePerformer, ProcedureFocalDevice
from .Encounter import Encounter, EncounterParticipant, EncounterDiagnosis, EncounterHospitalization
from .DocumentReference import DocumentReference
from .DiagnosticReport import DiagnosticReport
from .Immunization import Immunization
from .Condition import Condition
from .Condition import ConditionStage
from .Condition import ConditionEvidence
from .MedicationRequest import MedicationRequest
from .MedicationStatement import MedicationStatement
from .Medication import Medication
from .MedicationAdministration import MedicationAdministration
from .CareTeam import CareTeam
from .CarePlan import CarePlan
from .Organization import Organization
from .Practitioner import Practitioner, PractitionerQualification
from .Provenance import Provenance
from .SupplyDelivery import SupplyDelivery
from .PractitionerRole import PractitionerRole
from .AllergyIntolerance import AllergyIntolerance
from .ImagingStudy import ImagingStudy
from .Location import Location
from .Patient import Patient
from .Device import Device
from .ClinicalImpression import ClinicalImpression
from .Composition import Composition, CompositionEventType,CompositionRelatesTo,CompositionSection
from .SimpleFHIRStore import SimpleFHIRStore


from __future__ import annotations
import abc
from typing import TYPE_CHECKING, Optional, Tuple, Union
from pydantic.error_wrappers import ValidationError
from pydantic import HttpUrl, AnyUrl, parse_obj_as
from fhirkit.primitive_datatypes import literal, canonical

if TYPE_CHECKING:
    from .elements import Identifier, Reference
    from .Resource import Resource


class ExpansionError(Exception):
    pass


class ResourceNotFoundError(Exception):
    pass


class AbstractFHIRServer(abc.ABC):
    CURRENT_SERVER: Optional[AbstractFHIRServer] = None

    def __init__(self, base_url: Optional[Union[str, HttpUrl]] = None) -> None:
        self._base_url = parse_obj_as(AnyUrl, base_url) if base_url is not None else base_url

    @property
    def base_url(self):
        return self._base_url

    @abc.abstractmethod
    def get_resource_by_canonical(self, reference: Union[canonical, str]) -> "Resource":
        """Resolves a canonical reference into a resource or raises a ResourceNotFound exception.
        More info on canonical references: https://build.fhir.org/references.html#canonical"""
        # check if urn or absolute url
        # check version
        # check fragment
        pass

    @abc.abstractmethod
    def get_resource_by_id(self, resourceId: str, resourceType: Optional[str] = None):
        """Resolves a resourceType and id couple into a resource or raises a ResourceNotFound exception."""

    @abc.abstractmethod
    def get_resource_by_literal(self, reference: literal) -> "Resource":
        """Resolves a literal reference into a resource or raises a ResourceNotFound exception.
        More info on literal references: https://build.fhir.org/references.html#literal"""
        # check if urn
        # check if absolute url
        # check if relative url
        # check if fragment
        pass

    @abc.abstractmethod
    def get_resource_by_identifier(
        self,
        *,
        identifier: "Identifier",
        resourceType: str,
    ) -> "Resource":
        """Resolves an Identifier element into a resource or raises a ResourceNotFound exception."""
        pass

    def get_resource_by_reference(self, reference: "Reference") -> "Resource":
        """Resolves a Reference into a resource or raises a ResourceNotFound exception"""
        try:
            if reference.reference is not None:
                return self.get_resource_by_literal(reference.reference)
        except ResourceNotFoundError:
            pass

        try:
            if reference.identifier is not None and reference.type is not None:
                return self.get_resource_by_identifier(identifier=reference.identifier, resourceType=reference.type)
        except ResourceNotFoundError:
            pass
        raise ResourceNotFoundError(f"Could not resolve reference: {reference}")



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore
from pydantic import Field, validator
from typing import Optional, Sequence, Union, List
from fhirkit.choice_type import deterimine_choice_type, ChoiceType
from fhirkit.primitive_datatypes import dateTime, date
from fhirkit.Resource import DomainResource
from fhirkit.metadata_types import ContactDetail, ContactPoint
from fhirkit.elements import (
    CodeableConcept,
    HumanName,
    Identifier,
    Reference,
    Address,
    AdministrativeGender,
    Attachment,
    BackboneElement,
    Link
)


class PatientCommunication(BackboneElement):
    language: CodeableConcept = Field(
        None,
        title="The language which can be used to communicate with the patient about his or her health")
    preferred: Optional[bool] = Field(
        None,
        title="Language preference indicator")
class Patient(DomainResource):
    resourceType: Literal["Patient"] = Field(
        "Patient", 
        const=True)
    identifier: Optional[Sequence[Identifier]] = Field(
        None,
        title="An identifier for this patient")
    active: Optional[bool] = Field(
        None,
        title="Whether this patient record is in active use")
    name: Optional[Sequence[HumanName]] = Field(
        None,
        title="A name associated with the patient")
    telecom: Optional[Sequence[ContactPoint]] = Field(
        None,
        title="A contact detail for the individual")
    gender: Optional[AdministrativeGender] = Field(
        None,
        title="male | female | other | unknown")
    birthDate: Optional[date] = Field(
        None,
        title="The date of birth for the individual")
    deceasedBoolean: Optional[bool] = Field(
        None, 
        exclude=False)
    deceasedDateTime: Optional[dateTime] = Field(
        None, 
        exclude=False)
    address: Optional[Sequence[Address]] = Field(
        None,
        title="Addresses for the individual")
    maritalStatus: Optional[CodeableConcept] = Field(
        None,
        title="Marital (civil) status of a patient",
        valueset="http://hl7.org/fhir/ValueSet/marital-status")
    multipleBirthBoolean: Optional[bool] = Field(
        None, 
        exclude=True)
    multipleBirthInteger: Optional[int] = Field(
        None, 
        exclude=True)
    multipleBirth: Optional[Union[bool, int]] = ChoiceType(
        None,
        title="Whether patient is part of a multiple birth")
    photo: Optional[Sequence[Attachment]] = Field(
        None,
        title="Image of the patient")
    contact: Optional[Sequence[ContactDetail]] = Field(
        None,
        title="A contact party (e.g. guardian, partner, friend) for the patient")
    communication: Optional[Sequence[PatientCommunication]] = Field(
        None,
        title="A language which may be used to communicate with the patient about his or her health")
    generalPractitioner: Optional[Sequence[Reference]] = Field(
        None,
        enum_reference_types=["Organization","Practitioner","PractitionerRole"],
        title="Patient's nominated primary care provider")
    managingOrganization: Optional[Reference] = Field(
        None,
        enum_reference_types=["Organization"],
        title="Organization that is the custodian of the patient record")
    link: Optional[Sequence[Link]] = Field(
        None,
        title="Link to another patient resource that concerns the same actual patient")




try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type:ignore
from typing import Optional, Union, List, Sequence
from pydantic import Field, validator,AnyUrl
from fhirkit.choice_type import deterimine_choice_type
from fhirkit.Resource import DomainResource, ResourceWithMultiIdentifier
from fhirkit.elements import (
    CodeableConcept,
    Identifier,
    Period,
    Reference,
    Annotation,
    BackboneElement,
    Range
)
from fhirkit.primitive_datatypes import dateTime,URI
from fhirkit.Practitioner import Practitioner


ProcedureStatus = Literal[
    "preparation",
    "in-progress",
    "not-done",
    "on-hold",
    "stopped",
    "completed",
    "entered-in-error",
    "unknown",
]


class ProcedurePerformer(BackboneElement):
    function: Optional[CodeableConcept] = Field(
        None,
        title="Type of performance")
    actor: Reference = Field(
        None,
        enum_reference_types=["Practitioner","PractitionerRole","Organization","Patient","RelatedPerson","Device"],
        title="The reference to the practitioner")
    onBehalfOf: Optional[Reference] = Field(
        None,
        enum_reference_types=["Organization"],
        title="Organization the device or practitioner was acting for")

class ProcedureFocalDevice(BackboneElement):
    action: Optional[CodeableConcept] = Field(
        None,
        title="Kind of change to device")
    manipulated: Reference = Field(
        None,
        enum_reference_types=["Device"],
        title="Device that was changed")

class Procedure(DomainResource, ResourceWithMultiIdentifier):
    resourceType: Literal["Procedure"] = Field(
        "Procedure", 
        const=True)
    identifier: Optional[Sequence[Identifier]] = Field(
        None,
        title="External Identifiers for this procedure")
    instantiatesCanonical: Optional[List[AnyUrl]] = Field(
        None,
        title="Instantiates FHIR protocol or definition",
        enum_canonical_types=["PlanDefinition","ActivityDefinition","Measure","OperationDefinition","Questionnaire"])
    instantiatesUri: Optional[List[URI]] = Field(
        None)
    basedOn: Optional[List[Reference]] = Field(
        None,
        enum_reference_types=["CarePlan","ServiceRequest"],
        title="A request for this procedure")
    partOf: Optional[List[Reference]] = Field(
        None,
        enum_reference_types=["Procedure","Observation","MedicationAdministration"],
        title="Part of referenced event")
    status: ProcedureStatus = Field(
        None,
        title="preparation | in-progress | not-done | on-hold | stopped | completed | entered-in-error | unknown",
        valueset="http://hl7.org/fhir/ValueSet/event-status")
    statusReason: Optional[CodeableConcept] = Field(
        None,
        title="Reason for current status")
    category: Optional[CodeableConcept] = Field(
        None,
        title="Classification of the procedure")
    code: Optional[CodeableConcept] = Field(
        None,
        title="Identification of the procedure")
    subject: Reference = Field(
        None,
        enum_reference_types=["Patient","Group"],
        title="Who the procedure was performed on")
    encounter: Optional[Reference] = Field(
        None,
        enum_reference_types=["Encounter"],
        title="Encounter created as part of")
    #performedDateTime: Optional[dateTime] = Field(
    #    None, 
    #    exclude=False)
    #performedPeriod: Optional[Period] = Field(
    #    None, 
    #    exclude=False)
    #performedString: Optional[str] = Field(
    #    None, 
    #    exclude=False)
    #performedAge: Optional[int] = Field(
    #    None, 
    #    exclude=False)
    #performedRange: Optional[Range] = Field(
    #    None, 
    #    exclude=False)
    performed: Optional[Union[dateTime, Period, str, int]] = Field(
        None,
        title="When the procedure was performed")
    recorder: Optional[Reference] = Field(
        None,
        enum_reference_types=["Patient","RelatedPerson","Practitioner","PractitionerRole"],
        title="Who recorded the procedure")
    asserter: Optional[Practitioner] = Field(
        None,
        enum_reference_types=["Patient","RelatedPerson","Practitioner","PractitionerRole"],
        title="Person who asserts this procedure")
    performer: Optional[Sequence[ProcedurePerformer]] = Field(
        None,
        title="The people who performed the procedure")
    location: Optional[Reference] = Field(
        None,
        enum_reference_types=["Location"],
        title="Where the procedure happened")
    reasonCode: Optional[List[CodeableConcept]] = Field(
        None,
        title="Coded reason procedure performed")
    reasonReference: Optional[List[Reference]] = Field(
        None,
        enum_reference_types=["Condition","Observation","DiagnosticReport","DocumentReference"],
        title="The justification that the procedure was performed")
    bodySite: Optional[List[CodeableConcept]] = Field(
        None,
        title="Target body sites")
    complication: Optional[List[CodeableConcept]] = Field(
        None,
        title="Complication following the procedure")
    complicationDetail: Optional[List[Reference]] = Field(
        None,
        enum_reference_types=["Condition"])
    followUp: Optional[List[CodeableConcept]] = Field(
        None,
        title="Instructions for follow up")
    note: Optional[List[Annotation]] = Field(
        None,
        title="Additional information about the procedure")
    focalDevice: Optional[List[ProcedureFocalDevice]] = Field(
        None,
        title="Device that was changed")
    usedReference: Optional[List[Reference]] = Field(
        None,
        enum_reference_types=["Device","Medication","Substance"],
        title="Items used during procedure")
    usedCode: Optional[List[CodeableConcept]] = Field(
        None,
        title="Coded items used during the procedure")


import logging
from pathlib import Path
import resource
from typing import (
    Callable,
    Generator,
    Generic,
    List,
    Optional,
    Sequence,
    TypeVar,
    Union,
)
from uuid import uuid5, uuid4

from pydantic import HttpUrl, ValidationError, parse_file_as, parse_obj_as
from fhirkit.elements.elements import CodeableConcept, Coding, Reference, Identifier
from fhirkit.primitive_datatypes import (
    URI,
    AbsoluteURL,
    Code,
    RelativeURL,
    URN,
    canonical,
    literal,
)
from fhirkit.Bundle import Bundle
from fhirkit.Server import AbstractFHIRServer, ResourceNotFoundError
from fhirkit.TerminologyServer import AbstractFHIRTerminologyServer
from fhirkit.Resource import Resource, ResourceWithMultiIdentifier
from fhirkit.parse import AnyPatientResource, parse_json_as_resource
from tqdm import tqdm

LOGGER = logging.getLogger(__name__)
R = TypeVar("R", bound=Resource)


def traverse(
    path: Path, filter_expr: Optional[Callable[[Path], bool]] = None
) -> Generator[Path, None, None]:
    for childPath in path.iterdir():
        if childPath.is_dir():
            if filter_expr:
                yield from [
                    path
                    for path in traverse(childPath, filter_expr)
                    if filter_expr(path)
                ]
            else:
                yield from traverse(childPath)
        else:
            if filter_expr is None or filter_expr(childPath):
                yield childPath


class SimpleFHIRStore(Generic[R], AbstractFHIRTerminologyServer, AbstractFHIRServer):
    """Simple FHIR Server holds a list of resources in memory. It can be initialised based on bulk export directory."""

    def __init__(
        self,
        resources: Sequence[R] = [],
        references: Sequence[Reference] = [],
        base_url: Optional[Union[str, HttpUrl]] = None,
    ) -> None:
        self.resources = []
        self._resources = list(resources)
        self._references = list(references)
        super().__init__(base_url)

    def iter(self):
        """Dummy method because we can't expect all child classes to be iterable."""
        yield from self._resources

    def __iter__(self):
        return self.iter()

    def __len__(self):
        return len(self._resources)

    def get_resource_by_canonical(self, reference: Union[canonical, str]) -> "Resource":
        reference = parse_obj_as(canonical, reference)
        version = reference.version
        uri = reference.uri
        for r in self._resources:
            # skip resources that don't have url field or url doesn't match

            # AVR 20221025 hasattr and getattr is not a good practice, we should use structural typing with
            # a CanonicalResource as defined in https://build.fhir.org/canonicalresource.html and
            # use runtime typechecking with a CanonicaResource protocol
            if not hasattr(r, "url") or getattr(r, "url") != uri:
                continue

            if version is not None:

                # skip resources that don't have a version field or the version doesn't match
                if not hasattr(r, "version") or getattr(r, "version") != version:
                    continue

            return r

        raise ResourceNotFoundError(f"Couldn't resolve canonical reference {reference}")

    def get_resource_by_id(
        self, resourceId: str, resourceType: Optional[str] = None
    ) -> "Resource":
        for r in self._resources:
            if r.resourceType == resourceType and r.id == resourceId:
                return r
        raise ResourceNotFoundError(
            f"Couldn't find a {resourceType} resource with id={resourceId}"
        )

    def get_resource_by_literal(
        self, reference: Union[literal, str], resourceType: Optional[str] = None
    ) -> "Resource":
        reference = parse_obj_as(literal, reference)
        if isinstance(reference, AbsoluteURL):
            if self.base_url is not None:
                if reference.host != self.base_url.host:
                    reference_str = reference.build(
                        scheme=reference.scheme,
                        host=reference.host or "",
                    )
                    raise ResourceNotFoundError(
                        f"Can't resolve a resource with different base url. Expected {self.base_url} but received {reference_str}"
                    )
                    return

        if isinstance(reference, RelativeURL):
            return self.get_resource_by_id(
                reference.resourceId, resourceType=reference.resourceType
            )

        if isinstance(reference, URN):
            return self.get_resource_by_id(reference, resourceType=resourceType)

        raise ResourceNotFoundError(
            f"Couldn't resolve resource given literal uri: {reference}"
        )

    def get_resource_by_identifier(
        self, resourceType: str, identifier: "Identifier"
    ) -> "Resource":
        for r in self._resources:
            if r.resourceType != resourceType:
                continue

            if isinstance(r, ResourceWithMultiIdentifier):
                for r_identifier in r.identifier:
                    if r_identifier == identifier and r.resourceType == resourceType:
                        return r

        raise ResourceNotFoundError(
            f"Couldn't resolver {resourceType} resource with identifier={identifier}"
        )

    def __getitem__(self, key):
        return self.get_resource_by_literal(key)

    def filter(self, expr: Callable[[R], bool]):
        return SimpleFHIRStore(
            [r for r in self.iter() if expr(r)], base_url=self.base_url
        )

    def create_reference(
        self, resource: R, auto_save_in_store: bool = True
    ) -> Reference:
        if auto_save_in_store and resource.id is None:
            self.put_resource(resource)
        assert resource.id is not None, "Resource is not known in this store."
        reference=Reference(reference=f"{resource.resourceType}/{resource.id}")
        self._references.append(reference)
        return reference

    def put_resource(self, resource: R):
        ## TODO charlotte
        ## if resource with matching resource.id exists override
        ## if resource with matching reource.identifier exists raise Error
        literal_id = uuid4().urn
        resource.id = literal_id
        self.resources.append(resource)

    def valueset_expand(self, *args, **kwargs):
        return super().valueset_expand(*args, **kwargs)

    def valueset_validate_code(
        self,
        url: URI,
        code: Optional[Code] = None,
        display: Optional[str] = None,
        system: Optional[URI] = None,
        coding: Optional[Coding] = None,
        codeableConcept: Optional[CodeableConcept] = None,
    ) -> bool:
        return super().valueset_validate_code(
            url, code, display, system, coding, codeableConcept
        )

    def _repr_html_(self):
        return (
            """
        <div>
            <style scoped>
                .dataframe tbody tr th:only-of-type {
                    vertical-align: middle;
                }

                .dataframe tbody tr th {
                    vertical-align: top;
                }

                .dataframe thead th {
                    text-align: right;
                }
            </style>
            <table border="1" class="dataframe">
                <thead>
                    <tr style="text-align: right;">
                    <th></th>
                    <th>resourceType</th>
                    <th>id</th>
                    </tr>
                </thead>
            <tbody>
            """
            + "".join(
                [
                    f"<tr><th>{i}</th><td>{r.resourceType}</td><td>{r.id}</td></tr>"
                    for i, r in enumerate(self._resources[:10])
                ]
            )
            + "<tr><th>...</th><td>...</td><td>...</td></tr>"
            if len(self) > 10
            else ""
            + """
                </tbody>
                </table>
                <p>%d resources</p>
                </div>
                """
            % len(self)
        )

    @classmethod
    def bulk_import(cls, path: Path):
        """Load resources from a directory containing NDJSON files with FHIR resources as if they are exported from the FHIR Bulk export API."""
        resources = []
        if isinstance(path, str):
            path = Path(path)
        target_paths = list(traverse(path, lambda p: p.suffix == ".ndjson"))
        for fpath in tqdm(target_paths):
            for i, line in tqdm(
                enumerate(open(fpath, "r")), desc=str(path), leave=True
            ):
                try:
                    resources.append(parse_json_as_resource(line))
                except ValidationError:
                    LOGGER.warning(
                        "Couldn't parse line %d from '%s'",
                        i,
                        str(fpath.absolute()),
                        exc_info=True,
                    )
        return cls(resources)

    @classmethod
    def load_bundles(cls, path: Path, fail_when_invalid: bool = False):
        """Load resources from a directory containing JSON files with FHIR Bundle resources per Patients."""
        resources: List[R] = []
        if isinstance(path, str):
            path = Path(path)
        target_paths = list(traverse(path, lambda p: p.suffix == ".json"))
        for fpath in tqdm(target_paths):
            bundle: Bundle = parse_file_as(Bundle, fpath)
            for i, entry in tqdm(enumerate(bundle.entry), desc=str(fpath), leave=True):
                if entry.resource is None:
                    LOGGER.warning(
                        "Entry %d in Bundle with path %s has no resoruce.",
                        i,
                        str(fpath.absolute()),
                    )
                    continue
                try:
                    resources.append(
                        parse_obj_as(AnyPatientResource, entry.resource.dict())
                    )
                except ValidationError:
                    LOGGER.warning(
                        "Couldn't load entry %d (fullURL='%s') from '%s'",
                        i,
                        entry.fullUrl,
                        str(fpath.absolute()),
                        exc_info=True,
                    )
                    if fail_when_invalid:
                        raise StopIteration("Invalid resource.")
        return cls(resources)



from importlib.resources import Resource
from typing import List, Union
import typing


try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated  # type: ignore

from pydantic import Field, parse_obj_as
from fhirkit.Resource import RESOURCE_MODELS

AnyPatientResource = Annotated[
    Union[tuple(RESOURCE_MODELS)],
    Field(discriminator="resourceType"),
]


def parse_json_as_resource(json: str):
    return parse_obj_as(AnyPatientResource, json)



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore
from pydantic import Field


from fhirkit.Resource import DomainResource


class CarePlan(DomainResource):
    resourceType: Literal["CarePlan"] = Field("CarePlan", const=True)



from __future__ import annotations
import warnings

from fhirkit.elements.elements import Reference

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore
from typing import (
    AbstractSet,
    Any,
    List,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
    Generator,
)
from pydantic import Extra, Field, PrivateAttr
from fhirkit.BaseModel import BaseModel
from fhirkit.primitive_datatypes import URI, Code, Id, Instant, dateTime, date
from fhirkit.elements import (
    AbstractFHIRServer,
    BackboneElement,
    CodeableConcept,
    Element,
    Identifier,
    Narrative,
    Extension,
    Coding,
    Period,
    Range,
    UsageContext,
    Meta
)
from fhirkit.metadata_types import ContactDetail

InclusionExclusion = Union[AbstractSet[Union[int, str]], Mapping[Union[int, str], Any]]
AbstractSetIntStr = AbstractSet[Union[int, str]]
MappingIntStrAny = Mapping[Union[int, str], Any]

RESOURCE_MODELS = []

class Resource(BaseModel):

    resourceType: str
    id: Optional[str] = Field(None, repr=False)
    meta: Optional[Meta] = Field(None, repr=False)
    implicitRules: Optional[URI] = Field(None, repr=False)
    language: Optional[Code] = Field(None, repr=False)

    def __init_subclass__(cls) -> None:
        RESOURCE_MODELS.append(cls)

    def __str__(self) -> str:
        text = self.resourceType
        if self.id:
            text += f"/{self.id}"
        return text

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.allow

    def get_references(self):
        for k, v in self._iter():
            if isinstance(v, Reference):
                yield k, v

    def to_reference(self):
        return Reference(type=self.resourceType, reference=f"{self.resourceType}/{self.id}")


class DomainResource(Resource):
    text: Optional[Narrative] = Field(None, repr=False)
    contained: Sequence[Resource] = Field([], repr=False)
    extension: Sequence[Extension] = Field([], repr=False)
    modifierExtension: Sequence[Extension] = Field([], repr=False)

    def _repr_html_(self):
        if self.text:
            return self.text.div
        else:
            return repr(self)


class ResourceWithMultiIdentifier(Resource):
    identifier: Sequence[Identifier] = []


class CanonicalResource(DomainResource, ResourceWithMultiIdentifier):
    url: Optional[URI] = None
    version: Optional[str] = None
    name: Optional[str] = None
    title: Optional[str] = None
    status: Literal["draft", "active", "retired", "unknown"]
    experimental: Optional[bool] = None
    date: Optional[Union[date, dateTime]] = None
    publisher: Optional[str] = None
    contact: Sequence[ContactDetail] = []
    description: Optional[str] = None
    useContext: Sequence[UsageContext] = []
    jurisdiction: Sequence[CodeableConcept] = []
    purpose: Optional[str] = None
    copyright: Optional[str] = None


from __future__ import annotations
from functools import total_ordering
import itertools
from typing import TYPE_CHECKING, Any, Generic, Optional, Sequence, TypeVar, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

from pydantic import AnyUrl, Field, ConstrainedList

from fhirkit.BaseModel import BaseModel

if TYPE_CHECKING:
    from fhirkit.Resource import Resource
from fhirkit.primitive_datatypes import URI, Code, XHTML, Instant, dateTime, literal,Id
from fhirkit.Server import AbstractFHIRServer, ResourceNotFoundError


class Element(BaseModel):
    id: Optional[str] = None
    extension: Optional[Sequence[Extension]] = Field(default_factory=list, repr=False)


class Attachment(Element):
    contentType: Optional[Code]
    language: Optional[Code]
    # TODO how to type base64 encode bytes


class Narrative(Element):
    status: Literal["generated", "extensions", "additional", "required"]
    div: XHTML


class Extension(Element):
    url: URI
    value: Optional[Any] = None


class BackboneElement(Element):
    modifierExtension: Sequence[Extension] = Field(default_factory=list, repr=False)


Narrative.update_forward_refs()
BackboneElement.update_forward_refs()
Element.update_forward_refs()


class AbstractCoding(Element):
    """FHIR Terminology based model for concepts"""

    display: Optional[str] = None
    code: str
    system: Optional[URI] = None
    version: Optional[str] = None

    class Config:
        allow_mutation = False


@total_ordering
class Coding(AbstractCoding):
    system: URI
    userSelected: Optional[bool] = None

    def __repr__(self) -> str:
        if self.display:
            return f'"{self.display}" {self.system}|{self.code}'
        else:
            return f"{self.system}|{self.code}"

    def __str__(self) -> str:
        return self.display or f"{self.system}|{self.code}"

    @classmethod
    def from_fsh(cls, fsh: str) -> Coding:
        system, code = fsh.split("#")
        if "|" in system:
            system, version = system.split("|")
        else:
            version = None
        return Coding(system=system, version=version, code=code)

    def fsh(self, include_display: bool = False, include_version: bool = True):
        token: str = self.system
        if self.version is not None and include_version:
            token += "|" + self.version
        token += "#" + self.code
        if self.display is not None and include_display:
            token += ' "{self.display}"'
        return token

    def __eq__(self, other: object) -> bool:
        if isinstance(other, AbstractCoding):
            return self.system == other.system and self.code == other.code
        return False

    def __lt__(self, other: AbstractCoding) -> bool:
        return (self.system, self.code) < (other.system, other.code)

    def __hash__(self):
        return hash((self.system, self.code))


class CodeableConcept(BaseModel):
    """FHIR Terminology based model for CodeableConcepts"""

    coding: Sequence[Coding] = Field(
        default_factory=list,
        title="Code defined by a terminology system")
    text: Optional[str] = Field(
        None,
        title="Plain text representation of the concept")

    def __str__(self) -> str:
        return self.text or super().__str__()

    def __eq__(self, other: object) -> bool:

        if isinstance(other, Coding):
            result = any(other == c for c in self.coding)
            return result
        elif isinstance(other, CodeableConcept):
            return any(
                c1 == c2 for c1, c2 in itertools.product(self.coding, other.coding)
            )
        else:
            return False


CodeableConcept.update_forward_refs()

class Period(Element):
    start: Optional[dateTime] = Field(
        None,
        title="Start time with inclusive boundary")
    end: Optional[dateTime] = Field(
        None,
        title="End time with inclusive boundary")

class UnresolveableReference(ValueError):
    def __init__(self, reference: Reference, *args) -> None:
        self.reference = reference
        super().__init__(*args)

    def __str__(self) -> str:
        return super().__str__()


class Reference(Element):
    reference: Optional[str] = Field(
        None,
        title="Literal reference pointing to a resource with a URL",
        description="This can be an absolute URL or relative URL (implying the to use the ). Fragments can be used to point to contained resources",)
    type: Optional[URI] = Field(
        None,
        title="Type of the resource (e.g. Patient)")
    identifier: Optional[Identifier] = Field(
        None,
        title="Logical reference, when literal reference is not known")
    display: Optional[str] = Field(
        None,
        title="Text alternative for the resource")

    def __repr__(self) -> str:
        if self.display is not None:
            return self.display
        if self.type is not None and self.reference is not None:
            return self.type + "/" + self.reference
        return super().__repr__()

    def resolve(self, store: AbstractFHIRServer) -> "Resource":
        assert (
            self.identifier is not None or self.reference is not None
        ), "Only references with a valid `identifier` or `reference` attribute can be resolved."
        try:
            return store.get_resource_by_reference(self)
        except ResourceNotFoundError as exc:
            raise UnresolveableReference(reference=self) from exc

IdentifierCode = Literal["usual", "official", "temp", "secondary (If known)"]
class Identifier(Element): 
    use: Optional[IdentifierCode] = Field(
        None,
        title="usual | official | temp | secondary (If known)")
    type: Optional[CodeableConcept] = Field(
        None,
        title="Description of identifier")
    system: Optional[URI] = Field(
        None,
        title="The namespace for the identifier value")
    value: Optional[str] = Field(
        None,
        title="The value that is unique")
    period: Optional[Period] = Field(
        None,
        title="Time period when id is/was valid for use")
    assigner: Optional[Reference] = Field(
        None,
        enum_reference_types=["Organization"],
        title="Organization that issued id (may be just text)")

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, Identifier):
            return False
        return __o.system == self.system and __o.value == self.value


Identifier.update_forward_refs()
Reference.update_forward_refs()


class Quantity(Element):
    value: float
    comparator: Optional[Literal["<", "<=", ">=", ">"]]
    unit: str
    system: Optional[AnyUrl]
    code: Optional[Code]

    def __str__(self) -> str:
        return f"{self.value} {self.unit}"

    def unit_as_coding(self):
        if self.code is None:
            raise TypeError("Can't convert a unit without code to a coding.")
        return Coding(display=self.unit, code=self.code, system=self.system)

    def __repr__(self) -> str:
        rep = f"{self.value} {self.unit}"
        if self.comparator:
            return self.comparator + " " + rep
        return rep


class SimpleQuantity(Quantity):
    comparator: None = Field(..., const=True)


# TODO specify these derived quantities further
class Age(Quantity):
    pass


class Duration(Quantity):
    pass


class Range(Element):
    low: Optional[SimpleQuantity]
    high: Optional[SimpleQuantity]


class Ratio(Element):
    numerator: Optional[Quantity] = None
    denominator: Optional[Quantity] = None


class ContactPoint(Element):
    system: Optional[
        Literal["phone", "fax", "email", "pager", "url", "sms", "other"]
    ] = None
    value: Optional[str] = None
    use: Optional[Literal["home", "work", "temp", "old", "mobile"]] = None
    period: Optional[Period] = None


AdministrativeGender = Literal["male", "female", "other", "unknown"]

DaysOfWeek = Literal["mon", "tue", "wed", "thu", "fri", "sat", "sun"]

HumanNameUse = Literal[
    "usual",
    "official",
    "temp",
    "nickname",
    "anonymous",
    "old",
    "maiden",
]


class HumanName(Element):
    use: Optional[HumanNameUse] = Field(None, repr=True)
    text: Optional[str] = Field(None, repr=True)
    family: Optional[str] = Field(None, repr=True)
    given: Sequence[str] = Field([], repr=True)
    prefix: Sequence[str] = Field([], repr=True)
    suffix: Sequence[str] = Field([], repr=True)
    period: Optional[Period] = Field(None, repr=True)


AddressUse = Literal[
    "home",
    "work",
    "temp",
    "old",
    "billing - purpose of this address",
]

AddressType = Literal["postal", "physical", "both"]


class Address(Element):
    use: AddressUse = Field("home", repr=True)
    type: AddressType = Field("postal", repr=True)
    text: Optional[str] = Field(None, repr=True)
    line: Sequence[str] = Field([], repr=True)
    city: Optional[str] = Field(None, repr=True)
    district: Optional[str] = Field(None, repr=True)
    state: Optional[str] = Field(None, repr=True)
    postalCode: Optional[str] = Field(None, repr=True)
    country: Optional[str] = Field(None, repr=True)
    period: Optional[Period] = Field(None, exclude=True)


class Annotation(Element):
    author: Optional[Reference] = Field(None, exclude=True)
    time: Optional[dateTime] = Field(None, exclude=True)
    text: str = Field(...)


class SignatureType(ConstrainedList):
    item_type = Coding
    min_items = 1
    unique_items = True
    __args__ = (Coding,)


class Signature(Element):
    type: SignatureType
    when: Instant
    who: Reference
    onBehalfOf: Optional[Reference] = None
    targetFormat: Optional[Code] = None
    sigFormat: Optional[Code] = None
    data: Optional[bytes]

LinkType = Literal["replaced-by", "replaces", "refer", "seealso"]
class Link(Element):
    other: Reference = Field(
        None,
        enum_reference_types=["Patient","RelatedPerson"],)
    type: LinkType = Field(
        None,
        title="replaced-by | replaces | refer | seealso")

class Meta(Element):
    versionId: Optional[Id] = Field(
        None,
        title="Version specific identifier")
    lastUpdated: Optional[Instant] = Field(
        None,
        title="When the resource version last changed")
    source: Optional[URI] = Field(
        None,
        title="Identifies where the resource comes from")
    profile: Optional[Sequence[AnyUrl]] = Field(
        default_factory=list,
        enum_canonical_values=["StructureDefinition"],
        title="Profiles this resource claims to conform to")
    security: Optional[Sequence[Coding]] = Field(
        default_factory=list,
        title="Security Labels applied to this resource")
    tag: Optional[Sequence[Coding]] = Field(
        default_factory=list,
        title="Tags applied to this resource")


try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore
from typing import Optional, Sequence, List
from pydantic import Field
from fhirkit.Resource import DomainResource, ResourceWithMultiIdentifier
from fhirkit.elements import (
    CodeableConcept,
    Coding,
    Reference,
    Identifier,
    Period,
    Duration,
)
from fhirkit.elements.elements import BackboneElement


EncounterStatus = Literal[
    "planned",
    "arrived",
    "triaged",
    "in-progress",
    "onleave",
    "finished",
    "cancelled",
]

EncounterClass = Literal[
    "AMB",
    "EMER",
    "FLD",
    "HH",
    "IMP",
    "ACUTE",
    "NONAC",
    "OBSENC",
    "PRENC",
    "SS",
    "VR",
]

class EncounterStatusHistory(BackboneElement):
    status: EncounterStatus = Field(
        "planned",
        title="planned | arrived | triaged | in-progress | onleave | finished | cancelled")
    period: Optional[Period] = Field(
        None,
        title="The time that the episode was in the specified status")

class EncounterClassHistory(BackboneElement):
    class_: Coding = Field(
        Coding(code="AMB",system= "http://hl7.org/fhir/v3/ActCode"), 
        alias='class',
        title="inpatient | outpatient | ambulatory | emergency +")
    period: Period = Field(
        None,
        title="Time period during which the patient was in the specified class")

class EncounterParticipant(BackboneElement):
    type: Optional[List[CodeableConcept]] = Field(
        None,
        title="Role of participant in encounter",
        valueset="http://hl7.org/fhir/ValueSet/encounter-participant-type")
    period: Optional[Period] = Field(
        None,
        title="Period of time during the encounter that the participant participated")
    individual: Optional[Reference] = Field(
        None,
        enum_reference_types=["Practitioner","PractitionerRole","RelatedPerson"],
        title="Persons involved in the encounter other than the patient")

class EncounterDiagnosis(BackboneElement):
    condition: Reference = Field(
        None,
        enum_reference_types=["Condition"],
        title="The diagnosis or procedure relevant to the encounter")
    use: Optional[CodeableConcept] = Field(
        None,
        title="Role that this diagnosis has within the encounter (e.g. admission, billing, discharge …)",
        valueset="http://hl7.org/fhir/ValueSet/diagnosis-role")
    rank: Optional[int] = Field(
        None,
        title="Ranking of the diagnosis (for each role type)")

class EncounterHospitalization(BackboneElement):
    preAdmissionIdentifier: Optional[Identifier] = Field(
        None,
        title="Pre-admission identifier")
    origin: Optional[Reference] = Field(
        None,
        enum_reference_types=["Location","Organization"],
        title="The location/organization from which the patient came before admission")
    admitSource: Optional[CodeableConcept] = Field(
        None,
        title="From where patient was admitted (physician referral, transfer)",
        valueset="http://hl7.org/fhir/ValueSet/encounter-admit-source")
    reAdmission: Optional[CodeableConcept] = Field(
        None,
        title="The type of hospital re-admission that has occurred (if any). If the value is absent, then this is not identified as a readmission",
        valueset="http://terminology.hl7.org/ValueSet/v2-0092")
    destination: Optional[Reference] = Field(
        None,
        enum_reference_types=["Location","Organization"],
        title="Location/organization to which the patient is discharged")
    dischargeDisposition: Optional[CodeableConcept] = Field(
        None,
        title="Category or kind of location after discharge",
        valueset="http://hl7.org/fhir/ValueSet/encounter-discharge-disposition")

class EncounterLocation(BackboneElement):
    location: Reference = Field(
        None,
        enum_reference_types=["Location"],
        title="Location the encounter takes place")
    status: Optional[EncounterStatus] = Field(
        None,
        title="planned | active | reserved | completed")
    period: Optional[Period] = Field(
        None,
        title="Time period during which the patient was present at the location")

class Encounter(DomainResource, ResourceWithMultiIdentifier):
    resourceType: Literal["Encounter"] = Field(
        "Encounter", 
        const=True)
    identifier: Sequence[Identifier] = Field(
        None,
        title="Identifier(s) by which this encounter is known")
    status: EncounterStatus = Field(
        "planned",
        title="planned | arrived | triaged | in-progress | onleave | finished | cancelled")
    statusHistory: Optional[List[EncounterStatusHistory]] = Field(
        None,
        title="List of past encounter statuses")
    class_: Coding = Field(
        Coding(code="AMB",system= "http://hl7.org/fhir/v3/ActCode"), 
        alias='class',
        title="Classification of patient encounter")
    classHistory: Optional[List[EncounterClassHistory]] = Field(
        None,
        title="List of past encounter classes")
    type: Optional[List[CodeableConcept]] = Field(
        None,
        title="Specific type of encounter",
        valueset="http://hl7.org/fhir/ValueSet/encounter-type")
    serviceType: Optional[CodeableConcept] = Field(
        None,
        title="Specific type of service")
    priority: Optional[CodeableConcept] = Field(
        None,
        title="Indicates the urgency of the encounter",
        valueset="http://terminology.hl7.org/ValueSet/v3-ActPriority")
    subject: Optional[Reference] = Field(
        None,
        enum_reference_types=["Patient","Group"],
        title="The patient or group present at the encounter")
    episodeOfCare: Optional[List[Reference]] = Field(
        None,enum_reference_types=["EpisodeOfCare"],
        title="Episode(s) of care that this encounter should be recorded against")
    basedOn: Optional[List[Reference]] = Field(
        None,
        enum_reference_types=["ServiceRequest"],
        title="The ServiceRequest that initiated this encounter")
    participant:  Optional[EncounterParticipant] = Field(
        None,
        title="List of participants involved in the encounter")
    appointment: Optional[List[Reference]] = Field(
        None,
        enum_reference_types=["Appointment"],
        title="The appointment that scheduled this encounter")
    period: Optional[Period] = Field(
        None,
        title="The start and end time of the encounter")
    length: Optional[Duration] = Field(
        None,
        title="Quantity of time the encounter lasted (less time absent)")
    reasonCode: Optional[List[CodeableConcept]] = Field(
        None,
        title="Coded reason the encounter takes place",
        valueset="http://hl7.org/fhir/ValueSet/encounter-reason")
    reasonReference: Optional[List[Reference]] = Field(
        None,
        enum_reference_types=["Condition","Procedure","Observation","ImmunizationRecommendation"],
        title="Reason the encounter takes place (reference)")
    diagnosis: Optional[List[EncounterDiagnosis]] = Field(
        None,
        title="The list of diagnosis relevant to this encounter")
    account: Optional[List[Reference]] = Field(
        None,enum_reference_types=["Account"],
        title="The set of accounts that may be used for billing for this Encounter")
    hospitalization: Optional[EncounterHospitalization] = Field(
        None, 
        title="Details about the admission to a healthcare service")
    location: Optional[List[EncounterLocation]] = Field(
        None,
        title="List of locations where  the patient has been")
    serviceProvider: Optional[Reference] = Field(
        None,
        enum_reference_types=["Organization"],
        title="The organization (facility) responsible for this encounter")
    partOf: Optional[Reference] = Field(
        None,
        enum_reference_types=["Encounter"],
        title="Another Encounter this encounter is part of")


    




try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore
from typing import Sequence
from pydantic import Field


from fhirkit.Resource import DomainResource, ResourceWithMultiIdentifier
from fhirkit.elements.elements import Identifier


class AllergyIntolerance(DomainResource, ResourceWithMultiIdentifier):
    resourceType: Literal["AllergyIntolerance"] = Field(
        "AllergyIntolerance", const=True
    )
    identifier: Sequence[Identifier] = Field([])



from datetime import datetime  # this is important to have at the top

from typing import List, Optional, Sequence, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

from pydantic import Field
from fhirkit.choice_type import ChoiceType
from fhirkit.Resource import DomainResource
from fhirkit.elements import (
    CodeableConcept,
    Identifier,
    Period,
    Reference,
)

DiagnosticReportStatus = Literal[
    "registered",
    "partial",
    "preliminary"
    "final",
]

class DiagnosticReport(DomainResource):
    resourceType: Literal["DiagnosticReport"] = Field("DiagnosticReport", const=True)
    identifier: Optional[List[Identifier]] = Field(None)
    basedOn: Optional[List[Reference]] = Field(None)
    status: Optional[DiagnosticReportStatus] = Field(None)
    category: Optional[List[CodeableConcept]] = Field(None)
    code: Optional[CodeableConcept] = Field(None)
    subject: Optional[Reference] = Field(None)
    encounter: Optional[Reference] = Field(None)
    effectiveDateTime: Optional[datetime] = Field(None)
    effectivePeriod: Optional[Period] = Field(None)
    effective: Optional[Union[datetime, Period,str]] = ChoiceType(None)
    issued: Optional[datetime] = Field(None)
    performer: Optional[List[Reference]] = Field(None)
    resultsInterpreter: Optional[List[Reference]] = Field(None)
    specimen: Optional[List[Reference]] = Field(None)
    result: Optional[List[Reference]] = Field(None)
    conclusion: Optional[str] = Field(None)
    conclusionCode: Optional[List[CodeableConcept]] = Field(None)





try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore
from pydantic import Field
from typing import Optional, Sequence, List
from fhirkit.Resource import DomainResource, ResourceWithMultiIdentifier
from fhirkit.elements import (
    CodeableConcept,
    Identifier,
    ContactPoint,
    Address,
    Reference,
    HumanName,
    BackboneElement
)

class OrganizationContact(BackboneElement):
    purpose: Optional[CodeableConcept] = Field(
        None,
        title="The type of contact")
    name: Optional[HumanName] = Field(
        None,
        title="A name associated with the contact")
    telecom: Optional[Sequence[ContactPoint]] = Field(
        None,
        title="Contact details (telephone, email, etc.)  for a contact")
    address: Optional[Address] = Field(
        None,
        title="Visiting or postal addresses for the contact")

class Organization(DomainResource, ResourceWithMultiIdentifier):
    resourceType: Literal["Organization"] = Field(
        "Organization", 
        const=True)
    identifier: Optional[Sequence[Identifier]] = Field(
        None,
        title="An identifier for this patient")
    active: Optional[bool] = Field(
        None,
        title="Whether this organization's record is in active use")
    type: Optional[Sequence[CodeableConcept]] = Field(
        None,
        title="Kind of organization")
    name: Optional[str] = Field(
        None,
        title="Name used for the organization")
    alias: Optional[List[str]] = Field(
        None,
        title="A list of alternate names that the organization is known as, or was known as in the past")
    telecom: Optional[Sequence[ContactPoint]] = Field(
        None,
        title="A contact detail for the organization")
    address: Optional[Sequence[Address]] = Field(
        None,
        title="An address for the organization")
    partOf: Optional[Reference] = Field(
        None,
        enum_reference_types=["Organization"],
        title="The organization of which this organization forms a part")
    contact: Optional[Sequence[OrganizationContact]] = Field(
        None,
        title="Contact for the organization for a certain purpose")
    endpoint: Optional[Sequence[Reference]] = Field(
        None,
        enum_reference_types=["Endpoint"],
        title="Technical endpoints providing access to services operated for the organization")



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from pydantic import Field


from fhirkit.Resource import DomainResource


class MedicationRequest(DomainResource):
    resourceType: Literal["MedicationRequest"] = Field("MedicationRequest", const=True)




try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type:ignore
from typing import Optional, Union, List, Sequence
from pydantic import Field, validator
from fhirkit.choice_type import deterimine_choice_type, ChoiceType
from fhirkit.Resource import DomainResource, ResourceWithMultiIdentifier
from fhirkit.elements import (
    CodeableConcept,
    Identifier,
    Period,
    Reference,
    Annotation,
    BackboneElement,
)
from fhirkit.primitive_datatypes import dateTime
from fhirkit.Practitioner import Practitioner
from fhirkit.Organization import Organization

class CompositionEventType(BackboneElement):
    code: Optional[CodeableConcept] = Field(None, repr=True)
    period: Optional[Period] = Field(None, repr=True)
    detail: Optional[List[Reference]] = Field([], repr=True)

class CompositionRelatesTo(BackboneElement):
    code: Optional[CodeableConcept] = Field(None, repr=True)
    targetIdentifier: Optional[Identifier] = Field(None, repr=True)
    targetReference: Optional[Reference] = Field(None, repr=True)
    target: Optional[Union[Identifier, Reference]] = ChoiceType(None)
    
class CompositionSection(BackboneElement):
    title: str = Field(None, repr=True)
    code: Optional[CodeableConcept] = Field(None, repr=True)
    text: Optional[Annotation] = Field(None, repr=True)
    mode: Optional[CodeableConcept] = Field(None, repr=True)
    orderedBy: Optional[CodeableConcept] = Field(None, repr=True)
    entry: Optional[List[Reference]] = Field([], repr=True)
    emptyReason: Optional[CodeableConcept] = Field(None, repr=True)
    #section: Optional[List[CompositionSection]] = Field([], repr=True)

CompositionStatus = Literal["preliminary", "final", "amended", "entered-in-error"]

class Composition(DomainResource, ResourceWithMultiIdentifier):

    resourceType: Literal["Composition"] = Field("Composition", const=True)
    author: Optional[Reference] = Field(None, repr=True)
    category: Optional[CodeableConcept] = Field(None, repr=True)
    custodian: Optional[Reference] = Field(None, repr=True)
    date: Optional[dateTime] = Field(None, repr=True)
    encounter: Optional[Reference] = Field(None, repr=True)
    event: Optional[Sequence[CompositionEventType]] = Field([], repr=True)
    identifier: Optional[Sequence[Identifier]] = Field([], repr=True)
    relatesTo: Optional[Sequence[CompositionRelatesTo]] = Field([], repr=True)
    section: Optional[Sequence[CompositionSection]] = Field([], repr=True)
    status: CompositionStatus = Field("final", repr=True)
    subject: Optional[Reference] = Field(None, repr=True)
    title: str = Field(None, repr=True)
    type: Optional[CodeableConcept] = Field(None, repr=True)




from __future__ import annotations
from typing import (
    Any,
    Sequence,
)

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

from pydantic import Field
from fhirkit import r5
from fhirkit.utils.deprecated import deprecated


class CodeLookupError(KeyError):
    pass


@deprecated(reason="Use r5.CodeSystem instead")
class CSConceptDesignation(r5.CodeSystemConceptDesignation):
    pass


class CSConceptProperty(r5.CodeSystemConceptProperty):
    @property
    def value(
        self,
    ) -> (
        r5.code
        | r5.Coding
        | r5.string
        | r5.integer
        | r5.boolean
        | r5.dateTime
        | r5.decimal
    ):
        for prop in self.__fields__.keys():
            if prop.startswith("value"):
                v = getattr(self, prop)
                if v is not None:
                    return v
        raise RuntimeError("No value found in CSConceptProperty")

    def __str__(self) -> str:
        return str(self.code) + ": " + str(self.value)


class CSConcept(r5.CodeSystemConcept):
    property: Sequence[CSConceptProperty] = []
    concept: Sequence[CSConcept] = []

    def __getattribute__(self, __name: str) -> Any:
        try:
            return super().__getattribute__(__name)
        except AttributeError:
            for param in self.property:
                if param.code == __name:
                    return param.value
            raise

    def __str__(self) -> str:
        return (
            f"#{self.code} '{self.display}'"
            + "\nproperties: \n"
            + "\n".join(" " + str(p) for p in self.property if p.value is not None)
        )


class CSConceptLookup(CSConcept):
    name: str


class CodeSystem(r5.CodeSystem):
    resourceType: Literal["CodeSystem"] = Field("CodeSystem", const=True)


CodeSystem.update_forward_refs()



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore
from typing import Iterable, Sequence, Union, List
from pydantic import Field, validator
from fhirkit import r5
from fhirkit.utils.deprecated import deprecated
from fhirkit.elements import (
    CodeableConcept,
    Coding,
)


@deprecated(reason="Use r5.ValueSetComposeIncludeConceptDesignation instead")
class VSDesignation(r5.ValueSetComposeIncludeConceptDesignation):
    """This class is deprecated. Use r5.ValueSetComposeIncludeConceptDesignation instead."""

    pass


@deprecated(reason="Use r5.ValueSetComposeIncludeConcept instead")
class VSConcept(r5.ValueSetComposeIncludeConcept):
    """This class is deprecated. Use r5.ValueSetComposeIncludeConcept instead."""

    pass


ALLOWED_OPERATORS = [
    r5.code("="),
    r5.code("is-a"),
    r5.code("descendent-of"),
    r5.code("is-not-a"),
    r5.code("regex"),
    r5.code("in"),
    r5.code("not-in"),
    r5.code("generalizes"),
    r5.code("child-of"),
    r5.code("descendent-leaf"),
    r5.code("exists"),
]


class VSFilter(r5.ValueSetComposeIncludeFilter):
    """Select concepts by specifying a matching criterion based on the properties (including relationships) defined by the system, or on filters defined by the system. If multiple filters are specified within the include, they SHALL all be true."""

    op: r5.code = Field(default=...)
    "The kind of operation to perform as a part of the filter criteria."

    @validator("op")
    def operator_must_be_allowed(cls, v):
        if v not in ALLOWED_OPERATORS:
            raise ValueError(
                f"Operator {v.operator} is not allowed. Allowed operators are {ALLOWED_OPERATORS}"
            )
        return v


class VSInclude(r5.ValueSetComposeInclude):
    """Include one or more codes from a code system or other value set(s)."""

    filter: Sequence[VSFilter] = Field(default_factory=list)


class VSCompose(r5.ValueSetCompose):
    """A set of criteria that define the contents of the value set by including or excluding codes selected from the specified code system(s) that the value set draws from. This is also known as the Content Logical Definition (CLD)."""

    include: Sequence[VSInclude] = Field(default=...)
    "Include one or more codes from a code system or other value set(s)."
    exclude: Sequence[VSInclude] = Field(default_factory=list)
    "Exclude one or more codes from the value set based on code system filters and/or other value sets."


@deprecated(reason="Use r5.ValueSetExpansionContainsProperty instead")
class VSCodingProperty(r5.ValueSetExpansionContainsProperty):
    """This class is deprecated. Use r5.ValueSetExpansionContainsProperty instead."""

    pass


@deprecated(reason="Use r5.ValueSetExpansionContains instead")
class VSCodingWithDesignation(r5.ValueSetExpansionContains):
    """This class is deprecated. Use r5.ValueSetExpansionContains instead."""


@deprecated(reason="Use r5.ValueSetExpansion instead")
class VSExpansion(r5.ValueSetExpansion):
    """This class is deprecated. Use r5.ValueSetExpansion instead."""


class ValueSet(r5.ValueSet):
    resourceType: Literal["ValueSet"] = Field("ValueSet", const=True)
    compose: VSCompose | None = Field(default=None)
    "A set of criteria that define the contents of the value set by including or excluding codes selected from the specified code system(s) that the value set draws from. This is also known as the Content Logical Definition (CLD)."


class SimpleValueSet(ValueSet):
    status: r5.code = Field("active", const=True)
    expansion: VSExpansion

    def __init__(self, *args: VSCodingWithDesignation, **kwargs):
        if len(args) > 0:
            assert (
                "expansion" not in kwargs
            ), "When passing an iterable with concepts, `expansion` should be None."
            super().__init__(
                expansion=VSExpansion.parse_obj(
                    {"contains": [c.dict() for c in args], "total": len(args)}
                ),
                text=r5.Narrative(
                    status=r5.code("generated"),
                    div=r5.xhtml(
                        """
                <div>
                    <style scoped>
                        .dataframe tbody tr th:only-of-type {
                            vertical-align: middle;
                        }

                        .dataframe tbody tr th {
                            vertical-align: top;
                        }

                        .dataframe thead th {
                            text-align: right;
                        }
                    </style>
                    <table border="1" class="dataframe">
                        <thead>
                            <tr style="text-align: right;">
                            <th>code</th>
                            <th>display</th>
                            <th>system</th>
                            <th>version</th>
                            </tr>
                        </thead>
                        <tbody>"""
                        + "".join(
                            [
                                f"<tr><th>{c.code}</th><td>{c.display}</td><td>{c.system}</td><td>{c.version}</td></tr>"
                                for c in args
                            ]
                        )
                        + """
                        </tbody>
                    </table>
                </div>""",
                    ),
                ),
                **kwargs,
            )
        else:
            super().__init__(**kwargs)

    def append(
        self,
        code: VSCodingWithDesignation,
    ):
        assert (
            self.expansion is not None
        ), "`self.expansion` is None after initialisation with `self.init_expansion`"
        self.expansion.contains = list(self.expansion.contains) + [code]

    def extend(
        self,
        codes: Iterable[VSCodingWithDesignation],
    ):
        assert (
            self.expansion is not None
        ), "`self.expansion` is None after initialisation with `self.init_expansion`"
        self.expansion.contains = list(self.expansion.contains) + list(codes)

    def validate_code(self, code: Union[Coding, CodeableConcept]):
        if isinstance(code, CodeableConcept):
            return any(self.validate_code(c) for c in code.coding)
        elif isinstance(code, Coding):
            return any(c == code for c in self)
        else:
            return False


ValueSet.update_forward_refs()



from pydantic import Field, HttpUrl
from fhirkit.elements import Coding

class TiroCoding(Coding):
    system: HttpUrl = Field(default="http://tiro.health/fhir/CodeSystem/auto-generated")

    def __init__(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], str):
            code, display, *_ = (
                x.strip() for x in args[0].split("|")
            )  # break string in pieces
            super().__init__(code=code, display=display, **kwargs)
        else:
            super().__init__(*args, **kwargs)  # business as usual, Pydantic takes over



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type:ignore
from typing import Optional, Union, Sequence
from pydantic import Field, validator,AnyUrl
from fhirkit.choice_type import deterimine_choice_type, ChoiceType
from fhirkit.Resource import DomainResource, ResourceWithMultiIdentifier
from fhirkit.elements import (
    CodeableConcept,
    Identifier,
    Period,
    Reference,
    Age,
    Range,
    Period,
    BackboneElement,
    Annotation
)
from fhirkit.primitive_datatypes import dateTime
from fhirkit.Resource import DomainResource, ResourceWithMultiIdentifier
from fhirkit.elements.elements import Identifier

class ConditionStage(BackboneElement):
    summary: Optional[CodeableConcept] = Field(
        None,
        title="Simple summary (disease specific)")
    assessment: Optional[Sequence[Reference]] = Field(
        None,
        enum_reference_types=["ClinicalImpression","DiagnosticReport","Observation"],
        title="Formal record of assessment")
    type: Optional[CodeableConcept] = Field(
        None,
        title="Kind of staging")

class ConditionEvidence(BackboneElement):
    code: Optional[Sequence[CodeableConcept]] = Field(
        None,
        title="Manifestation/symptom")
    detail: Optional[Sequence[Reference]] = Field(
        None,
        enum_reference_types=["Any"],
        title="Supporting information found elsewhere")


class Condition(DomainResource, ResourceWithMultiIdentifier):
    resourceType: Literal["Condition"] = Field(
        "Condition", 
        const=True)
    identifier: Sequence[Identifier] = Field(
        None,
        title="External Ids for this condition")
    clinicalStatus: Optional[CodeableConcept] = Field(
        None,
        title="active | recurrence | relapse | inactive | remission | resolved",
        valueset="http://hl7.org/fhir/ValueSet/condition-clinical")
    verificationStatus: Optional[CodeableConcept] = Field(
        None,
        title="unconfirmed | provisional | differential | confirmed | refuted | entered-in-error",
        valueset="http://hl7.org/fhir/ValueSet/condition-ver-status")
    category: Optional[Sequence[CodeableConcept]] = Field(
        None,
        title="problem-list-item | encounter-diagnosis",
        valueset="http://hl7.org/fhir/ValueSet/condition-category")
    severity: Optional[CodeableConcept] = Field(
        None,
        title="Subjective severity of condition",
        valueset="http://hl7.org/fhir/ValueSet/condition-severity")
    code: Optional[CodeableConcept] = Field(
        None,
        title="Identification of the condition, problem or diagnosis")
    bodySite: Optional[Sequence[CodeableConcept]] = Field(
        None,
        title="Anatomical location, if relevant")
    subject: Optional[Reference] = Field(
        None,
        enum_reference_types=["Patient","Group"],
        title="Who has the condition?")
    encounter: Optional[Reference] = Field(
        None,
        enum_reference_types=["Encounter"],
        title="Encounter created as part of")
    onsetDateTime: Optional[dateTime] = Field(
        None, 
        exclude=True)
    onsetAge: Optional[Age] = Field(
        None, 
        exclude=True)
    onsetPeriod: Optional[Period] = Field(
        None, 
        exclude=True)
    onsetRange: Optional[Range] = Field(
        None, 
        exclude=True)
    onsetString: Optional[str] = Field(
        None, 
        exclude=True)
    onset: Optional[Union[dateTime, Age, Period, Range, str]] = ChoiceType(
        None,
        title="Estimated or actual date, date-time, or age")
    abatementDateTime: Optional[dateTime] = Field(
        None, 
        exclude=True)
    abatementAge: Optional[Age] = Field(
        None, 
        exclude=True)
    abatementPeriod: Optional[Period] = Field(
        None, 
        exclude=True)
    abatementRange: Optional[Range] = Field(
        None, 
        exclude=True)
    abatementString: Optional[str] = Field(
        None, 
        exclude=True)
    abatement: Optional[Union[dateTime, Age, bool, Period, Range, str]] = ChoiceType(
        None,
        title="When in resolution/remission")
    recordedDate: Optional[dateTime] = Field(
        None,
        title="Date record was first recorded")
    recorder: Optional[Reference] = Field(
        None,
        enum_reference_types=["Practitioner","PractitionerRole","Patient","RelatedPerson"],
        title="Who recorded the condition?")
    asserter: Optional[Reference] = Field(
        None,
        enum_reference_types=["Practitioner","PractitionerRole","Patient","RelatedPerson"],
        title="Person who asserts this condition")
    stage: Optional[Sequence[ConditionStage]] = Field(
        None,
        title="Stage/grade, usually assessed formally")
    evidence: Optional[Sequence[ConditionEvidence]] = Field(
        None,
        title="Supporting evidence")
    note: Optional[Sequence[Annotation]] = Field(
        None,
        title="Additional information about the Condition")


    @validator("onset", pre=True, always=True, allow_reuse=True)
    def validate_performed(cls, v, values, field):
        return deterimine_choice_type(
            cls,
            v,
            values,
            field,
        )
    @validator("abatement", pre=True, always=True, allow_reuse=True)
    def validate_performed(cls, v, values, field):
        return deterimine_choice_type(
            cls,
            v,
            values,
            field,
        )



try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from pydantic import Field


from fhirkit.Resource import DomainResource
from typing import Optional, Union, List, Sequence
from pydantic import Field, validator
from fhirkit.choice_type import deterimine_choice_type, ChoiceType
from fhirkit.Resource import DomainResource
from fhirkit.elements import CodeableConcept, Identifier, Period, Reference, Annotation, BackboneElement
from fhirkit.elements.elements import Address, ContactPoint
from fhirkit.primitive_datatypes import dateTime
from fhirkit.Practitioner import Practitioner
from fhirkit.Organization import Organization


LocationStatus = Literal[
    "active",
    "suspended",
    "inactive",
]
LocationMode = Literal[
    "instance",
    "kind",
]
class Location(DomainResource):
    resourceType: Literal["Location"] = Field("Location", const=True)
    identifier: Sequence[Identifier] = Field([], repr=True)
    status: Optional[LocationStatus] = Field("active", repr=True)
    name: Optional[str] = Field(None, repr=True)
    alias: Optional[List[str]] = Field(None, repr=True)
    description: Optional[str] = Field(None, repr=True)
    mode: Optional[LocationStatus] = Field(None, repr=True)
    type: Optional[List[CodeableConcept]] = Field(None, repr=True)
    telecom: Optional[List[ContactPoint]] = Field(None, repr=True)
    address: Optional[Address] = Field(None, repr=True)
    physicalType: Optional[CodeableConcept] = Field(None, repr=True)
    managingOrganization: Optional[Organization] = Field(None, repr=True)




try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore
from pydantic import Field
from typing import Optional, Sequence
from fhirkit import Reference
from fhirkit.Resource import DomainResource
from fhirkit.elements import (
    dateTime,
    HumanName,
    Identifier,
    ContactPoint,
    Address,
    AdministrativeGender,
    Attachment,
    Period,
    BackboneElement,
    CodeableConcept
)

class PractitionerQualification(BackboneElement):
    identifier: Optional[Sequence[Identifier]] = Field(
        None,
        title="An identifier for this qualification for the practitioner")
    code: CodeableConcept = Field(
        None,
        title="Coded representation of the qualification")
    period: Optional[Period] = Field(
        None,
        title="Period during which the qualification is valid")
    issuer: Optional[Reference] = Field(
        None,
        enum_reference_types=["Organization"],
        title="Organization that regulates and issues the qualification")
        
class Practitioner(DomainResource):
    resourceType: Literal["Practitioner"] = Field(
        "Practitioner", 
        const=True)
    identifier: Optional[Sequence[Identifier]] = Field(
        None,
        title="External Identifiers for this procedure")    
    active: Optional[bool] = Field(
        None,
        title="Whether this practitioner's record is in active use")
    name: Optional[Sequence[HumanName]] = Field(
        None,
        title="The name(s) associated with the practitioner")
    telecom: Optional[Sequence[ContactPoint]] = Field(
        None,
        title="A contact detail for the practitioner (that apply to all roles)")
    address: Optional[Sequence[Address]] = Field(
        None,
        title="Address(es) of the practitioner that are not role specific (typically home address)")
    gender: Optional[AdministrativeGender] = Field(
        None,
        title="male | female | other | unknown")
    birthDate: Optional[dateTime] = Field(
        None,
        title="The date of birth for the practitioner")
    photo: Optional[Attachment] = Field(
        None,
        title="Image of the person")
    qualification: Optional[Sequence["PractitionerQualification"]] = Field(
        None,
        title="Certification, licenses, or training pertaining to the provision of care")
    communication: Optional[Sequence[CodeableConcept]] = Field(
        None,
        title="A language the practitioner is able to use in patient communication")
